const config = require('../config.json');
const logger = require('./logger.js');
const api = require('./api.js');
const blinky = require('./blinky.js');
const http = require('http');

const ID = require('./constants.js').VARIABLE_ID;

var server;

async function start() {
    if (!server) {
        server = http.createServer();
    }
    server.listen(config.server.port, config.server.listen).on('listening', function () {
        logger.debug('server listening on ' + config.server.listen + ':' + config.server.port + '...');
        blinky.setActive(true);
        handleRequests();
    });
}

function handleRequests() {
    server.on('request', function (request, response) {
        request.timestamp = new Date().getTime();
        if (request.url.length > 1 && request.url.endsWith('/')) {
            request.url = request.url.substring(0, request.url.length - 1);
        }
        let id;
        let endpointName = request.url;
        let separatorIndex = request.url.lastIndexOf('/');
        if (separatorIndex > 0) {
            id = request.url.substring(separatorIndex + 1, request.url.length);
            if (!isNaN(id)) {
                endpointName = endpointName.substring(0, separatorIndex + 1) + ID;
            }
        }
        var endpoint = api.getEndpoints().get(endpointName);
        if (endpoint === undefined) {
            endRequest(
                request,
                response,
                'endpoint \'' + request.url + '\' not defined',
                501
            );
            return;
        }
        endpoint = endpoint[request.method];
        if (endpoint === undefined || endpoint.method === undefined) {
            endRequest(
                request,
                response,
                'endpoint \'' + request.url + '\' does not have any handlers for ' + request.method + ' requests',
                405
            );
            return;
        }
        getRequestParams(request)
            .then(function (params) {
                return endpoint.method(id, params);
            })
            .then(function (result) {
                endRequest(request, response, result);
            })
            .catch(function (err, code) {
                if (code == undefined) {
                    code = 500;
                }
                endRequest(request, response, err, code);
            });
    });
}

function endRequest(request, response, msg, code) {
    if (code == undefined) {
        code = 200;
    }
    var object = {
        request: request,
        code: code
    }
    if (msg != undefined) {
        if (msg instanceof Error) {
            object.data = msg.toString();
        } else if (msg instanceof Object) {
            object.data = JSON.stringify(msg);
        } else {
            object.data = msg;
        }
    }
    response.setHeader("Access-Control-Allow-Origin", "*");
    response.setHeader("Access-Control-Allow-Headers", "Origin, X-Requested-With, Content-Type, Accept");
    response.writeHead(object.code);
    response.end(object.data);
    logger.http(object);
}

function getRequestParams(request) {
    return new Promise(function (resolve, reject) {
        var params = "";
        request.on("data", function (data) {
            params += data;
        });
        request.on("end", function () {
            if (params == undefined || params.length == 0) {
                return resolve();
            }
            return resolve(new URLSearchParams(params));
        });
    });
}

module.exports = {
    start
}